package com.oracle.psr.nettybug;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionStage;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.Invocation;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedHashMap;
import javax.ws.rs.core.Response;
import org.glassfish.jersey.client.ClientConfig;
import org.glassfish.jersey.client.ClientProperties;
import org.glassfish.jersey.netty.connector.NettyConnectorProvider;

public class NettyBug extends Thread {

    public static AtomicInteger count = new AtomicInteger();
    public static String url;
    public static int nLoops;

    private static Client client;

    public static void main(String[] args) throws InterruptedException {
        url = args[0];
        int nThreads = Integer.parseInt(args[1]);
        nLoops = Integer.parseInt(args[2]);
        initClient();
        Thread[] threads = new Thread[nThreads];
        for (int i = 0; i < nThreads; i++) {
            threads[i] = new NettyBug();
            threads[i].start();
        }

        for (int i = 0; i < nThreads; i++) {
            threads[i].join();
        }
        System.out.println("Processed calls: " + count);
        System.exit(0);
    }

    private static void initClient() {
        ClientConfig defaultConfig = new ClientConfig();
        defaultConfig.property(ClientProperties.CONNECT_TIMEOUT, 10 * 1000);
        defaultConfig.property(ClientProperties.READ_TIMEOUT, 10 * 1000);
        defaultConfig.connectorProvider(new NettyConnectorProvider());
        client = ClientBuilder.newBuilder()
                .withConfig(defaultConfig)
                .build();
    }

    public void doCall() {
        CompletableFuture<Response> cf = invokeResponse().toCompletableFuture()
                .whenComplete((rsp, t) -> {
                    if (t != null) {
                        System.out.println(Thread.currentThread() + " async complete. Caught exception " + t);
                    }
                })
                .handle((rsp, t) -> {
                    if (rsp != null) {
                        rsp.readEntity(String.class);
                    } else {
                        System.out.println(Thread.currentThread().getName() + " response is null");
                    }
                    return rsp;
                }).exceptionally(t -> {
                    System.out.println("async complete. completed exceptionally " + t);
                    throw new RuntimeException(t);
        });

        try {
            cf.get();
            System.out.println("Done call " + count.incrementAndGet());
        } catch (InterruptedException | ExecutionException ex) {
            Logger.getLogger(NettyBug.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private static CompletionStage<Response> invokeResponse() {
        WebTarget target = client.target(url);
        MultivaluedHashMap hdrs = new MultivaluedHashMap<>();
        /*
        List a1 = new ArrayList();
        a1.add("1234");
        List a2 = new ArrayList();
        a2.add("true");
        hdrs.put("x-span-id", a1);
        hdrs.put("x-oic-debug", a2);
        //String jsonPayload = "{\"fname\":\"foo\", \"lname\":\"bar\"}";
         */
        StringBuilder sb = new StringBuilder("{");
        for (int i = 0; i < 10000; i++) {
            sb.append("\"fname\":\"foo\", \"lname\":\"bar\"");
        }
        sb.append("}");
        String jsonPayload = sb.toString();
        Invocation.Builder builder = ((WebTarget) target).request().headers(hdrs);
        return builder.rx().method("POST", Entity.entity(jsonPayload, MediaType.APPLICATION_JSON_TYPE));
    }

    @Override
    public void run() {
        for (int i = 0; i < nLoops; i++) {
            try {
                doCall();
            } catch (Throwable t) {
                t.printStackTrace();
            }
        }
    }
}
